% This is the main program to do the three-stage learning using Word2Vec inputs
% Goal: train and test the model on 79 semeval relations.
% The output results are saved in the ./rst/SemEval/training20, 
% output 1: learned relational representations are saved as mat files including posterior means and covariance matrix).
% output 2: summary results including the correlation with human typicality ratings.


clear all;
tic; 
rng(1); % set random seed

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Input data and parameters %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

input = 'word2vec'; 
datapath = '.\data\SemEval\Phase2AnswersScaled\';

% set # of training examples
num_train_pairs_pos =  20; 	% number of positive training examples for comparative relations
outputrstpath = ['.\rst\SemEval\training' num2str(num_train_pairs_pos) '\'];
numpairperrel = 1;  % select # of word pair per relation to serve as negative examples for learning other relations

% model parameters
alphaval = .5;   % alpha value in stage-2 logistic regression with elastic net regularization. alpha = 0 is ridge regression with L2 regularization; alpha = 0 is lasso regression with L1 regularization
use_hyperprior = true; %false;%     % whether to use hyperprior in stage-3 Bayes learning
max_iter = 500;                 % maximum number of iterations in stage-3 Bayes learning for variational method
criterion = 0.00001;            % criterion for convergence for stage-3 Bayes learning for variational method

% Input-specific parameters
num_sel_dims = 300;    % # of feature dimensions for word vectors       

% input dataset: SemEval with 79 relations in 10 types
datasetname = 'SemEval';
filerel = {'1a','1b','1c','1d','1e','2a','2b','2c','2d','2e','2f','2g','2h','2i','2j',...
    '3a','3b','3c','3d','3e','3f','3g','3h','4a','4b','4c','4d','4e','4f','4g','4h',...
    '5a','5b','5c','5d','5e','5f','5g','5h','5i',...
    '6a','6b','6c','6d','6e','6f','6g','6h',...
    '7a','7b','7c','7d','7e','7f','7g','7h',...
    '8a','8b','8c','8d','8e','8f','8g','8h',...
    '9a','9b','9c','9d','9e','9f','9g','9h','9i',...
    '10a','10b','10c','10d','10e','10f'};%
feature_names = filerel;
typelab = {1:5, 6:15,16:23,24:31,32:40,41:48,49:56,57:64,65:73,74:79};

learnrellabel = 1:length(filerel);
num_features_to_learn = length(learnrellabel);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get input data for pre-processing %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

numfeats = num_sel_dims;
num_pair_feats = numfeats * 2;

%%%%%%%%%%%%%%%%%%%%%%%
% Run the simulations %
%%%%%%%%%%%%%%%%%%%%%%%

% construct negative example pool including the few pairs with top typicality ratings for each relation
all_topneg_vecs = []; all_topneg_names={};
for relation_to_learn = 1:length(filerel)
    %% load word vec
    datatoppair = [];
    load([datapath  'Phase2Answers-' filerel{relation_to_learn} '.txt.mat'],'allData'); 
    wordtemp = cell2mat(allData(1:2*numpairperrel,2:301));
    
    for si = 1:numpairperrel
        all_topneg_vecsrel(si,:) = [wordtemp((si-1)*2+1,:) wordtemp((si-1)*2+2,:)];
        all_topneg_namesrel{si,1} = strcat(allData{(si-1)*2+1,1}, ':', allData{(si-1)*2+2,1});
    end;
    
    all_topneg_vecs = [all_topneg_vecs;all_topneg_vecsrel];
    all_topneg_names((relation_to_learn-1)*numpairperrel+1:relation_to_learn*numpairperrel,1) = all_topneg_namesrel ;
end;

% get the rel category labels
for ti = 1:length(typelab)
    all_rellabel(typelab{ti}) = ti;
end;
all_topneg_rellabel = kron(all_rellabel,ones(1,numpairperrel));

for relation_to_learn = learnrellabel
    
    %% load human ratings
    filename = [datapath  'Phase2Answers-' filerel{relation_to_learn} '.txt'];
    fid = fopen( filename);
    data = textscan(fid, '%f %q','headerlines', 11);
    fclose(fid);

    humrating{relation_to_learn} = data{1,1};
    wordpairs = length(humrating{relation_to_learn}); % # of word-pairs from the SemEv dataset.


    %% load word vec
    datavalprot = [];
    load([datapath  'Phase2Answers-' filerel{relation_to_learn} '.txt.mat'],'allData'); 
    if size(allData,1)~=2*wordpairs   % there are some extra lines in the original mat files
        warning('more word pairs than human ratings. check the raw data file\n'); 
        allDatatemp = allData;
        allData = [];
        allData = allDatatemp(1:2*wordpairs,:);
        save([datapath  'Phase2Answers-' filerel{relation_to_learn} '.txt.mat'],'allData'); 
    end;

    datavalprot=cell2mat(allData(1:2*wordpairs,2:301));
    Gpairnum(relation_to_learn) = size(datavalprot,1)/2;  %wordpairs;%

    % generate positive pairs
    for si = 1:Gpairnum(relation_to_learn)
        all_pos_vecs(si,:) = [datavalprot((si-1)*2+1,:) datavalprot((si-1)*2+2,:)];
        all_pos_names{si,1} = strcat(allData{(si-1)*2+1,1}, ':', allData{(si-1)*2+2,1});
    end;
    

    feature_name = feature_names{relation_to_learn};
    fprintf('\nLearning about %s:', feature_name);
    
   
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Create the entire pool of possible training and test items %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    tot_num_pos_pairs = Gpairnum(relation_to_learn);
                
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Constucting training and testing sets
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Select positive training examples
    train_pos_inds = 1:num_train_pairs_pos; %randsample(tot_num_pos_pairs, num_train_pairs_pos);% %$$$$ %
    train_pos_vecs = all_pos_vecs(train_pos_inds, :);
    train_pos_names = all_pos_names(train_pos_inds, :);

    % generate negative pairs
    % use the top word pairs from other relations
    [~,indtemp]=find(all_topneg_rellabel~=all_rellabel(relation_to_learn));
    train_neg_vecs = all_topneg_vecs(indtemp,:);
    train_neg_names = all_topneg_names(indtemp);

    % Combine the training examples
    num_train_pairs_neg = size(train_neg_vecs,1);
    all_train_data = [train_pos_vecs; train_neg_vecs];
    all_train_labels = [ones(num_train_pairs_pos, 1); 0 * ones(num_train_pairs_neg, 1)];


    % Test all pairs in the target relation including training pairs
    test_pos_inds = 1:tot_num_pos_pairs;%setdiff(1:tot_num_pos_pairs, train_pos_inds);
    test_pos_vecs = all_pos_vecs(test_pos_inds, :);
    test_pos_names = all_pos_names(test_pos_inds, :);

    test_data = [test_pos_vecs];
    test_data_labels = humrating{relation_to_learn}>0; % find pairs that humans give positive typicality ratings
    test_data_names = [test_pos_names];
    num_test_pairs = size(test_data,1);
     
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Learning stage 1: feature augmentation      %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
    % construct difference vector bewteen any word pairs instantiate a relation
    data_diff = []; train_data_diff=[]; posdata_diff=[]; coeff=[]; mu=[];
    data_diff = (all_train_data(:, 1:num_sel_dims)-all_train_data(:, num_sel_dims+1:end));

    % concatenate ranked difference vector
    for ti = 1:size(all_train_data,1)
        temp1 = data_diff(ti, 1:num_sel_dims);
        [sorttemp1,sortindx(ti,:)] = sort(temp1);

        train_data_diff(ti,:) = [temp1,sorttemp1];
    end;
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Learning stage 2: feature selection using logistic regression with Elastic Net regularization. 
    % Input: concatenated difference vectors including the raw differece and ranked difference vectors
    % Output: imptdimindx, selected feature dimensions (including the raw and the ranked feature dimensions)
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

    X = train_data_diff;
    Y = all_train_labels;

    [B,FitInfo] = lassoglm(X,Y,'binomial','Alpha',alphaval,'MaxIter',10^5,'NumLambda',50,'CV','resubstitution');
    [~,indx]=min(FitInfo.Deviance); 
    % FitInfo.Lambda(indx)

    B0 = B(:,indx);
    [imptdimindx,temp] = find(B0 ~= 0);
    imptdimvalue = B0(imptdimindx);
    cnst = FitInfo.Intercept(indx);
    B1 = [cnst;B0];   

    % decision threshold
    threshold = .5; 

    % get accuracy for training dataset
    predsL = glmval(B1,X,'logit');
    rspL = predsL>threshold;        
    accuracyLi(relation_to_learn,1) = sum(Y==rspL)/length(Y); 

    % test new word pairs, and generate the concatanated difference vectors
    t_data_diff = []; test_data_diff = []; t_scoreall=[];
    t_data_diff = ( test_data(:, 1:num_sel_dims)-test_data(:, num_sel_dims+1:end));          
    for ti = 1:size(t_data_diff,1)
        temp1 = t_data_diff(ti, 1:num_sel_dims);
        [sorttemp1,sortindxt(ti,:)] = sort(temp1);

        test_data_diff(ti,:) = [temp1,sorttemp1];        
    end;

    preds = glmval(B1,test_data_diff,'logit');

    rsp = preds>threshold;        
    seldimi(relation_to_learn,1) = length(imptdimindx);  % selected feature dimensions
    accuracyi(relation_to_learn,1) = sum(test_data_labels==rsp)/length(test_data_labels); 
    corrPindxi(relation_to_learn,1) = corr(preds,  humrating{relation_to_learn},'type','Pearson');
    corrSindxi(relation_to_learn,1) = corr(preds,  humrating{relation_to_learn},'type','Spearman');

    indxhumpos=humrating{relation_to_learn}>0;
    corrSposindxi(relation_to_learn,1) = corr(preds(indxhumpos),  humrating{relation_to_learn}(indxhumpos),'type','Spearman');

    % generate selected feacture vetors of word pairs for Stage-3 Bayes learning
    train_data2raw=[]; test_data2raw=[]; train_data2=[]; test_data2=[]; train_data2label = [];
    for ti = 1:size(all_train_data,1)
         train_data2raw(ti,:) = [all_train_data(ti,1:numfeats) all_train_data(ti,sortindx(ti,:)) all_train_data(ti,numfeats+1:end) all_train_data(ti,numfeats+sortindx(ti,:))];

        if all_train_labels(ti)==1
            train_data2label(ti,1) = 1;
        else
            train_data2label(ti,1) = -1;
        end;
    end;
    
    % feature augmentation
    for ti = 1:size(test_data,1)
         test_data2raw(ti,:) = [test_data(ti,1:numfeats) test_data(ti,sortindxt(ti,:)) test_data(ti,numfeats+1:end) test_data(ti,numfeats+sortindxt(ti,:))];
    end;    

    totalobjfdim = size(test_data2raw,2); 
    % feature selection
    selobjfdim = [imptdimindx; numfeats*2+imptdimindx];
    train_data2 = [train_data2raw(:,selobjfdim) ];
    test_data2 = [test_data2raw(:,selobjfdim) ];       
    num_pair_feats = size(train_data2,2);

        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Learning stage 3:  Bayesian learning using selected features and empirical weight prior  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % centering
    dim_means = mean(train_data2,1);
    train_data2 = train_data2 - repmat(dim_means, size(train_data2,1), 1);
    test_data2 = test_data2 - repmat(dim_means, size(test_data2,1), 1);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Obtain prior distribution of weights %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Empirical prior: contrast based prior
    mu_prior = [imptdimvalue; -imptdimvalue];%

    %%%%%%%%%%%%%%%%%%%%%%
    % Learn the relation: Bayesian logistic model
    %%%%%%%%%%%%%%%%%%%%%%
    % use_hyperprior
    if use_hyperprior
        a0 = 1;
        b0 = 5;
        [mu_post, sigma_post, sigma_inv_post, logdetV, E_a, L, predprob, ...
            final_iter] = bayes_logit_fit_nonzero_mean(train_data2, ...
            train_data2label, mu_prior, a0, b0, max_iter, criterion);
    else  % without hyperprior    
        sigma_prior = eye(num_pair_feats) * 5;
            [mu_post, sigma_post, sigma_inv_post] = train_batch(train_data2, ...
                train_data2label, mu_prior, sigma_prior, max_iter, criterion);
    end;
    % Save posterior mu and covariance matrices
    mumat = [mu_prior mu_post];
    sigma_post = (sigma_post + sigma_post')/2;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Evaluate performance for test pairs and save results 
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    pred_prob_pos=[];
    for test = 1 : num_test_pairs
        pair_vec_pos = test_data2(test, :);
        pred_prob_pos(test,1) = compute_score(pair_vec_pos', 1, mu_post, sigma_inv_post, max_iter, criterion);
        %  fprintf('%d %s %f\n',test,char(test_pos_names{test,:}), pred_prob_pos(test,1) );
    end; 


    rsp2 = pred_prob_pos>threshold;        
    accuracyi2(relation_to_learn,1) = sum(test_data_labels==rsp2)/length(test_data_labels); 
    accuracyi2pos(relation_to_learn,1) = sum(rsp2(indxhumpos)==1)/sum(indxhumpos); 
    corrPindxi2(relation_to_learn,1) = corr(pred_prob_pos,  humrating{relation_to_learn},'type','Pearson');
    corrSindxi2(relation_to_learn,1) = corr(pred_prob_pos,  humrating{relation_to_learn},'type','Spearman');
    corrSposindxi2(relation_to_learn,1) = corr(pred_prob_pos(indxhumpos),  humrating{relation_to_learn}(indxhumpos),'type','Spearman');
     
        
    %% load golden pair to compute the corr using word2vec method
    W2VcalCorr;

    
    % display the summary results for each relation
    fprintf('#seldim = %f,w2vcorrS = %f, trainAcc = %f, testAcc = %f, corrS = %f,  BARTaccu = %f, BARTposaccu = %f, BARTcorrS = %f\n',seldimi(relation_to_learn,1), ...,
        corrval_w2v(relation_to_learn,2), accuracyLi(relation_to_learn,1),accuracyi(relation_to_learn,1),...
        corrSindxi(relation_to_learn,1), accuracyi2(relation_to_learn,1), accuracyi2pos(relation_to_learn,1),...
        corrSindxi2(relation_to_learn,1)); 

    % save the learning results in the output folder
    rst(relation_to_learn,:) = [relation_to_learn,corrval_w2v(relation_to_learn,2), corrval_w2vPos(relation_to_learn,2),(seldimi(relation_to_learn,1)), ...,
        (accuracyLi(relation_to_learn,1)),(accuracyi(relation_to_learn,1)), (corrSindxi(relation_to_learn,1)), (corrSposindxi(relation_to_learn,1)),...
        (accuracyi2(relation_to_learn,1)), (accuracyi2pos(relation_to_learn,1)),(corrSindxi2(relation_to_learn,1)),(corrSposindxi2(relation_to_learn,1))];
        
    save([outputrstpath sprintf('%s_mumat.mat', feature_names{relation_to_learn})], 'mumat');
    save([outputrstpath sprintf('%s_sigma.mat', feature_names{relation_to_learn})], 'sigma_post');
    save([outputrstpath sprintf('%s_rsttemp.mat', feature_names{relation_to_learn})],'imptdimindx','selobjfdim','dim_means','pred_prob_pos','rst','num_train_pairs_pos','totalobjfdim');% 
end  %relation_to_learn ends

% display the result in the command window for the ten types
for i = 1:length(typelab)
    labtemp = typelab{i};
    rstreltype(i,1) = i; 
    for j = 2:size(rst,2)
        rstreltype(i,j) = nanmean(rst(labtemp,j)); 
    end;
end;
rstreltype

avgrstreltyp = mean(rstreltype);

save([outputrstpath 'rstsummary.mat'], 'rst','rstreltype','avgrstreltyp');
toc;

% Test with UCLA veb analogy set
TestVebAnalogyfunc(outputrstpath,outputrstpath);
